#  vim: set fileencoding=utf-8 :

# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import unittest

from ykman.scancodes import KEYBOARD_LAYOUT, encode


class TestScanMap(unittest.TestCase):
    def test_us_layout(self):
        self.assertEqual(b"\x04", encode("a", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x05", encode("b", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x06", encode("c", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x07", encode("d", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x08", encode("e", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x09", encode("f", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x0a", encode("g", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x0b", encode("h", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x0c", encode("i", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x0d", encode("j", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x0e", encode("k", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x0f", encode("l", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x10", encode("m", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x11", encode("n", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x12", encode("o", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x13", encode("p", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x14", encode("q", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x15", encode("r", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x16", encode("s", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x17", encode("t", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x18", encode("u", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x19", encode("v", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x1a", encode("w", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x1b", encode("x", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x1c", encode("y", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x1d", encode("z", KEYBOARD_LAYOUT.US))

        self.assertEqual(b"\x84", encode("A", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x85", encode("B", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x86", encode("C", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x87", encode("D", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x88", encode("E", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x89", encode("F", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x8a", encode("G", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x8b", encode("H", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x8c", encode("I", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x8d", encode("J", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x8e", encode("K", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x8f", encode("L", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x90", encode("M", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x91", encode("N", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x92", encode("O", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x93", encode("P", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x94", encode("Q", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x95", encode("R", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x96", encode("S", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x97", encode("T", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x98", encode("U", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x99", encode("V", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x9a", encode("W", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x9b", encode("X", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x9c", encode("Y", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x9d", encode("Z", KEYBOARD_LAYOUT.US))

        self.assertEqual(b"\x27", encode("0", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x1e", encode("1", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x1f", encode("2", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x20", encode("3", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x21", encode("4", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x22", encode("5", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x23", encode("6", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x24", encode("7", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x25", encode("8", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x26", encode("9", KEYBOARD_LAYOUT.US))

        self.assertEqual(b"\x2b", encode("\t", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x28", encode("\n", KEYBOARD_LAYOUT.US))

        self.assertEqual(b"\x9e", encode("!", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb4", encode('"', KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa0", encode("#", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa1", encode("$", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa2", encode("%", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa4", encode("&", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x34", encode("'", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa6", encode("(", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa7", encode(")", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa5", encode("*", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xae", encode("+", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x35", encode("`", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x36", encode(",", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x2d", encode("-", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x37", encode(".", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x38", encode("/", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb3", encode(":", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x33", encode(";", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb6", encode("<", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x2e", encode("=", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb7", encode(">", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb8", encode("?", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x9f", encode("@", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x2f", encode("[", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x32", encode("\\", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\x30", encode("]", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xa3", encode("^", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xad", encode("_", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xaf", encode("{", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb0", encode("}", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb2", encode("|", KEYBOARD_LAYOUT.US))
        self.assertEqual(b"\xb5", encode("~", KEYBOARD_LAYOUT.US))

        self.assertEqual(b"\x04\x05\x06", encode("abc", KEYBOARD_LAYOUT.US))
        with self.assertRaises(ValueError):
            encode("ö")

    def test_de_layout(self):
        self.assertEqual(b"\x04", encode("a", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x05", encode("b", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x06", encode("c", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x07", encode("d", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x08", encode("e", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x09", encode("f", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x0a", encode("g", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x0b", encode("h", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x0c", encode("i", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x0d", encode("j", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x0e", encode("k", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x0f", encode("l", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x10", encode("m", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x11", encode("n", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x12", encode("o", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x13", encode("p", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x14", encode("q", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x15", encode("r", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x16", encode("s", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x17", encode("t", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x18", encode("u", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x19", encode("v", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x1a", encode("w", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x1b", encode("x", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x1d", encode("y", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x1c", encode("z", KEYBOARD_LAYOUT.DE))

        self.assertEqual(b"\x84", encode("A", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x85", encode("B", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x86", encode("C", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x87", encode("D", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x88", encode("E", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x89", encode("F", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x8a", encode("G", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x8b", encode("H", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x8c", encode("I", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x8d", encode("J", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x8e", encode("K", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x8f", encode("L", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x90", encode("M", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x91", encode("N", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x92", encode("O", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x93", encode("P", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x94", encode("Q", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x95", encode("R", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x96", encode("S", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x97", encode("T", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x98", encode("U", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x99", encode("V", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x9a", encode("W", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x9b", encode("X", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x9d", encode("Y", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x9c", encode("Z", KEYBOARD_LAYOUT.DE))

        self.assertEqual(b"\x27", encode("0", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x1e", encode("1", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x1f", encode("2", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x20", encode("3", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x21", encode("4", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x22", encode("5", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x23", encode("6", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x24", encode("7", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x25", encode("8", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x26", encode("9", KEYBOARD_LAYOUT.DE))

        self.assertEqual(b"\x2b", encode("\t", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x28", encode("\n", KEYBOARD_LAYOUT.DE))

        self.assertEqual(b"\x32", encode("#", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x30", encode("+", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x36", encode(",", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x38", encode("-", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x64", encode("<", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x35", encode("^", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x2c", encode(" ", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x2e", encode("´", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x2d", encode("ß", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x34", encode("ä", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x33", encode("ö", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x2f", encode("ü", KEYBOARD_LAYOUT.DE))

        self.assertEqual(b"\x9e", encode("!", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\x9f", encode('"', KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa1", encode("$", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa2", encode("%", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa3", encode("&", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb2", encode("'", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa5", encode("(", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa6", encode(")", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb0", encode("*", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa4", encode("/", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb7", encode(":", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb6", encode(";", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa7", encode("=", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xe4", encode(">", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xad", encode("?", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb8", encode("_", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xad", encode("`", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xa0", encode("§", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb4", encode("Ä", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xb3", encode("Ö", KEYBOARD_LAYOUT.DE))
        self.assertEqual(b"\xaf", encode("Ü", KEYBOARD_LAYOUT.DE))

        self.assertEqual(b"\xb4\xb3\xaf", encode("ÄÖÜ", KEYBOARD_LAYOUT.DE))
        with self.assertRaises(ValueError):
            encode("@", KEYBOARD_LAYOUT.DE)

    def test_norman_layout(self):
        self.assertEqual(b"\x04", encode("a", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x05", encode("b", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x06", encode("c", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x08", encode("d", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x07", encode("e", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x15", encode("f", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x0a", encode("g", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x33", encode("h", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x0e", encode("i", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x1c", encode("j", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x17", encode("k", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x12", encode("l", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x10", encode("m", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x0d", encode("n", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x0f", encode("o", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x11", encode("p", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x14", encode("q", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x0c", encode("r", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x16", encode("s", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x09", encode("t", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x18", encode("u", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x19", encode("v", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x1a", encode("w", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x1b", encode("x", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x0b", encode("y", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x1d", encode("z", KEYBOARD_LAYOUT.NORMAN))

        self.assertEqual(b"\x84", encode("A", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x85", encode("B", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x86", encode("C", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x88", encode("D", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x87", encode("E", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x95", encode("F", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x8a", encode("G", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xb3", encode("H", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x8e", encode("I", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x9c", encode("J", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x97", encode("K", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x92", encode("L", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x90", encode("M", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x8d", encode("N", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x8f", encode("O", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x91", encode("P", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x94", encode("Q", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x8c", encode("R", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x96", encode("S", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x89", encode("T", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x98", encode("U", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x99", encode("V", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x9a", encode("W", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x9b", encode("X", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x8b", encode("Y", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x9d", encode("Z", KEYBOARD_LAYOUT.NORMAN))

        self.assertEqual(b"\x27", encode("0", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x1e", encode("1", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x1f", encode("2", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x20", encode("3", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x21", encode("4", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x22", encode("5", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x23", encode("6", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x24", encode("7", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x25", encode("8", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x26", encode("9", KEYBOARD_LAYOUT.NORMAN))

        self.assertEqual(b"\x2b", encode("\t", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x28", encode("\n", KEYBOARD_LAYOUT.NORMAN))

        self.assertEqual(b"\xa0", encode("#", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xae", encode("+", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x36", encode(",", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"-", encode("-", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xb6", encode("<", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa3", encode("^", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x2c", encode(" ", KEYBOARD_LAYOUT.NORMAN))

        self.assertEqual(b"\x9e", encode("!", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xb4", encode('"', KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa1", encode("$", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa2", encode("%", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa4", encode("&", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"4", encode("'", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa6", encode("(", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa7", encode(")", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xa5", encode("*", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"8", encode("/", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xb3", encode(":", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\x13", encode(";", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b".", encode("=", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xb7", encode(">", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xb8", encode("?", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"\xad", encode("_", KEYBOARD_LAYOUT.NORMAN))
        self.assertEqual(b"5", encode("`", KEYBOARD_LAYOUT.NORMAN))

        self.assertEqual(b"\x04\x05\x06", encode("abc", KEYBOARD_LAYOUT.NORMAN))
        with self.assertRaises(ValueError):
            encode("ö")



# Copyright (c) 2020 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import sys
from threading import Event
from typing import Callable

from yubikit.core.otp import OtpConnection
from yubikit.support import read_info

from ..base import REINSERT_STATUS, CancelledException
from .base import OtpYubiKeyDevice
from .fido import list_ctap_devices

__all__ = [
    "list_otp_devices",
    "list_ctap_devices",
]
logger = logging.getLogger(__name__)


if sys.platform == "linux":
    from .linux import list_devices
elif sys.platform == "win32":
    from .windows import list_devices
elif sys.platform == "darwin":
    from .macos import list_devices
elif sys.platform == "freebsd":
    from .freebsd import list_devices
else:

    def list_devices() -> list:
        raise NotImplementedError("OTP HID support is not implemented on this platform")


list_otp_devices: Callable[[], list[OtpYubiKeyDevice]] = list_devices


def _otp_reinsert(
    self: OtpYubiKeyDevice, reinsert_cb: Callable[[REINSERT_STATUS], None], event: Event
) -> None:
    removed_state = None
    with self.open_connection(OtpConnection) as conn:
        info = read_info(conn, self.pid)

    reinsert_cb(REINSERT_STATUS.REMOVE)
    logger.debug(f"Waiting for removal of device {self.path}")
    while not event.wait(0.5):
        keys = list_otp_devices()
        present = {k.path for k in keys}
        if removed_state is None:
            if self.path not in present:
                logger.debug(f"Removed! {self.path}")
                reinsert_cb(REINSERT_STATUS.REINSERT)
                removed_state = present
        else:
            added = present - removed_state
            if len(added) == 1:
                dev_fp = next(iter(added))  # Path may have changed
                logger.debug(f"Inserted! {dev_fp}")
                key = next(k for k in keys if k.path == dev_fp)
                # Update path
                self.path = key.path
                with self.open_connection(OtpConnection) as conn:
                    info2 = read_info(conn, self.pid)
                if info.serial != info2.serial or info.version != info2.version:
                    raise ValueError("Reinserted YubiKey does not match the original")
                return
            elif len(added) > 1:
                raise ValueError("Multiple YubiKeys inserted")

    raise CancelledException()


# Patch the reinsert method to the OtpYubiKeyDevice class for the correct backend
setattr(OtpYubiKeyDevice, "_do_reinsert", _otp_reinsert)



# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import random
import struct
from datetime import datetime
from typing import Iterable

from yubikit.core.otp import modhex_encode
from yubikit.oath import parse_b32_key
from yubikit.yubiotp import YubiOtpSession

from .scancodes import KEYBOARD_LAYOUT

logger = logging.getLogger(__name__)


def is_in_fips_mode(session: YubiOtpSession) -> bool:
    """Check if the OTP application of a FIPS YubiKey is in FIPS approved mode.

    :param session: The YubiOTP session.
    """
    return session.backend.send_and_receive(0x14, b"", 1) == b"\1"  # type: ignore


DEFAULT_PW_CHAR_BLOCKLIST = ["\t", "\n", " "]


def generate_static_pw(
    length: int,
    keyboard_layout: KEYBOARD_LAYOUT = KEYBOARD_LAYOUT.MODHEX,
    blocklist: Iterable[str] = DEFAULT_PW_CHAR_BLOCKLIST,
) -> str:
    """Generate a random password.

    :param length: The length of the password.
    :param keyboard_layout: The keyboard layout.
    :param blocklist: The list of characters to block.
    """
    chars = [k for k in keyboard_layout.value.keys() if k not in blocklist]
    sr = random.SystemRandom()
    return "".join([sr.choice(chars) for _ in range(length)])


def parse_oath_key(val: str) -> bytes:
    """Parse a secret key encoded as either Hex or Base32.

    :param val: The secret key.
    """
    try:
        return bytes.fromhex(val)
    except ValueError:
        return parse_b32_key(val)


def format_oath_code(response: bytes, digits: int = 6) -> str:
    """Format an OATH code from a hash response.

    :param response: The response.
    :param digits: The number of digits in the OATH code.
    """
    offs = response[-1] & 0xF
    code = struct.unpack_from(">I", response[offs:])[0] & 0x7FFFFFFF
    return ("%%0%dd" % digits) % (code % 10**digits)


def time_challenge(timestamp: int, period: int = 30) -> bytes:
    """Format a HMAC-SHA1 challenge based on an OATH timestamp and period.

    :param timestamp: The timestamp.
    :param period: The period.
    """
    return struct.pack(">q", int(timestamp // period))


def format_csv(
    serial: int,
    public_id: bytes,
    private_id: bytes,
    key: bytes,
    access_code: bytes | None = None,
    timestamp: datetime | None = None,
) -> str:
    """Produce a CSV line in the "Yubico" format.

    :param serial: The serial number.
    :param public_id: The public ID.
    :param private_id: The private ID.
    :param key: The secret key.
    :param access_code: The access code.
    """
    ts = timestamp or datetime.now()
    return ",".join(
        [
            str(serial),
            modhex_encode(public_id),
            private_id.hex(),
            key.hex(),
            access_code.hex() if access_code else "",
            ts.isoformat(timespec="seconds"),
            "",  # Add trailing comma
        ]
    )



from yubikit.core import Tlv, Oid, bytes2int, int2bytes
import pytest

OID_TESTS = {
    "1.2.840.10045.3.1.7": b"\x2a\x86\x48\xce\x3d\x03\x01\x07",  # SECP256R1
    "1.3.132.0.10": b"\x2b\x81\x04\x00\x0a",  # SECP256K1
    "1.3.132.0.34": b"\x2b\x81\x04\x00\x22",  # SECP384R1
    "1.3.132.0.35": b"\x2b\x81\x04\x00\x23",  # SECP521R1
    "1.3.36.3.3.2.8.1.1.7": b"\x2b\x24\x03\x03\x02\x08\x01\x01\x07",  # BrainpoolP256R1
    "1.3.36.3.3.2.8.1.1.11": b"\x2b\x24\x03\x03\x02\x08\x01\x01\x0b",  # BrainpoolP384R1
    "1.3.36.3.3.2.8.1.1.13": b"\x2b\x24\x03\x03\x02\x08\x01\x01\x0d",  # BrainpoolP512R1
    "1.3.6.1.4.1.3029.1.5.1": b"\x2b\x06\x01\x04\x01\x97\x55\x01\x05\x01",  # X25519
    "1.3.6.1.4.1.11591.15.1": b"\x2b\x06\x01\x04\x01\xda\x47\x0f\x01",  # Ed25519
}


@pytest.mark.parametrize("oid_str, oid_bytes", OID_TESTS.items())
def test_oid_conversion(oid_str, oid_bytes):
    oid = Oid.from_string(oid_str)
    assert oid == Oid(oid_bytes)
    assert oid.dotted_string == oid_str
    assert bytes(oid) == oid_bytes

    # Test round-trip conversion
    assert Oid(bytes(oid)) == oid
    assert Oid.from_string(oid.dotted_string) == oid

    # Test that the bytes representation matches the expected bytes
    assert bytes(oid) == oid_bytes


@pytest.mark.parametrize("value", (b"", None))
def test_empty_tlv(value):
    # Test that an empty Tlv can be created and serialized
    tlv = Tlv(0x01, value)
    assert tlv.tag == 0x01
    assert tlv.length == 0
    assert tlv.value == b""
    assert bytes(tlv) == b"\x01\x00"


# Some sample TLV data for testing
TLV_TESTS = {
    "0100": Tlv(0x01, b""),
    "020101": Tlv(0x02, b"\x01"),
    "03020203": Tlv(0x03, b"\x02\x03"),
    "0403020101": Tlv(0x04, Tlv(0x02, b"\x01")),
}


@pytest.mark.parametrize("data, expected_tlv", TLV_TESTS.items())
def test_tlv_parsing(data, expected_tlv):
    data = bytes.fromhex(data)

    # Test parsing TLV data
    tlv = Tlv(data)
    assert tlv == expected_tlv

    # Test that the serialized TLV matches the original data
    assert bytes(tlv) == data

    # Test that unpacking the TLV gives the correct tag, length, and value
    assert tlv.tag == expected_tlv.tag
    assert tlv.length == expected_tlv.length
    assert tlv.value == expected_tlv.value


@pytest.mark.parametrize("ln", (0, 1, 0x7F, 0x80, 0xFF, 0x100, 0xFFFF, 0x10000))
def test_tlv_length_boundries(ln):
    data = b"\xcc" * ln
    tlv = Tlv(0x01, data)
    assert tlv.tag == 0x01
    assert tlv.length == ln
    assert tlv.value == data
    assert tlv == Tlv(tlv)


def test_tlv_multibyte_tag():
    # Test that multi-byte tags are handled correctly
    tlv = Tlv(0x7F49, b"test")
    assert tlv.tag == 0x7F49
    assert tlv.length == 4
    assert tlv.value == b"test"
    assert bytes(tlv) == b"\x7f\x49\x04test"

    # Test that the TLV can be parsed back correctly
    parsed_tlv = Tlv(bytes(tlv))
    assert parsed_tlv == tlv

    # Test invalid multi-byte tag
    with pytest.raises(ValueError):
        Tlv(0xFFFF, b"invalid")


@pytest.mark.parametrize("n", (0, 1, 0xFF, 0x100, 0xFFFF, 0x10000))
def test_int2bytes(n):
    assert bytes2int(int2bytes(n)) == n


def test_int2bytes_zero():
    assert int2bytes(0) == b"\x00"



#  vim: set fileencoding=utf-8 :

import pytest

from yubikit.oath import (
    HASH_ALGORITHM,
    OATH_TYPE,
    CredentialData,
    _derive_key,
    _format_cred_id,
    _parse_cred_id,
)


@pytest.mark.parametrize(
    ("raw", "issuer", "name", "period"),
    [
        (b"20/Issuer:name", "Issuer", "name", 20),
        (b"weird/Issuer:name", "weird/Issuer", "name", 30),
        (b"Issuer:name", "Issuer", "name", 30),
        (b"20/name", None, "name", 20),
        (b"name", None, "name", 30),
    ],
)
def test_parse_cred_id(raw, issuer, name, period):
    parsed_issuer, parsed_name, parsed_period = _parse_cred_id(raw, OATH_TYPE.TOTP)
    assert (parsed_issuer, parsed_name, parsed_period) == (issuer, name, period)


@pytest.mark.parametrize(
    ("issuer", "name", "period", "expected"),
    [
        (None, "name", None, b"name"),
        ("Issuer", "name", None, b"Issuer:name"),
        ("Issuer", "name", 20, b"20/Issuer:name"),
        ("Issuer", "name", 30, b"Issuer:name"),
        (None, "name", 20, b"20/name"),
    ],
)
def test_format_cred_id(issuer, name, period, expected):
    kwargs = {}
    if period is not None:
        kwargs["period"] = period
    assert _format_cred_id(issuer, name, OATH_TYPE.TOTP, **kwargs) == expected


@pytest.mark.parametrize(
    ("salt", "password", "expected"),
    [
        (
            b"\0" * 8,
            "foobar",
            b"\xb0}\xa1\xe7\xde\x87\xf8\x9a\x87\xa2\xb5\x98\xea\xa2\x18\x8c",
        ),
        (
            b"12345678",
            "Hallå världen!",
            b"\xda\x81\x8ek,\xf0\xa2\xd0\xbf\x19\xb3\xdd\xd3K\x83\xf5",
        ),
        (
            b"saltsalt",
            "Ťᶒśƫ ᵽĥřӓşḛ",
            b"\xf3\xdf\xa7\x81T\xc8\x102\x99E\xfb\xc4\xb55\xe57",
        ),
    ],
)
def test_derive_key(salt, password, expected):
    assert _derive_key(salt, password) == expected


@pytest.mark.parametrize(
    ("uri", "expected"),
    [
        ("otpauth://totp/account?secret=abba", None),
        ("otpauth://totp/account?secret=abba&issuer=Test", "Test"),
        ("otpauth://totp/Test:account?secret=abba", "Test"),
        ("otpauth://totp/TestA:account?secret=abba&issuer=TestB", "TestB"),
    ],
)
def test_parse_uri_issuer(uri, expected):
    assert CredentialData.parse_uri(uri).issuer == expected


def test_parse_uri_full_payload():
    data = CredentialData.parse_uri(
        "otpauth://totp/Issuer:account"
        "?secret=abba&issuer=Issuer"
        "&algorithm=SHA256&digits=7"
        "&period=20&counter=5"
    )
    assert data.secret == b"\0B"
    assert data.issuer == "Issuer"
    assert data.name == "account"
    assert data.oath_type == OATH_TYPE.TOTP
    assert data.hash_algorithm == HASH_ALGORITHM.SHA256
    assert data.digits == 7
    assert data.period == 20
    assert data.counter == 5



# Copyright (c) 2023 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os

import click
from cryptography.hazmat.primitives import serialization

from yubikit.core import TRANSPORT
from yubikit.core.smartcard import SW, ApduError, SmartCardConnection
from yubikit.hsmauth import (
    ALGORITHM,
    CREDENTIAL_PASSWORD_LEN,
    MANAGEMENT_KEY_LEN,
    HsmAuthSession,
    InvalidPinError,
)
from yubikit.management import CAPABILITY

from ..hsmauth import (
    generate_random_management_key,
    get_hsmauth_info,
)
from ..util import InvalidPasswordError, parse_private_key
from .util import (
    CliFail,
    click_callback,
    click_force_option,
    click_format_option,
    click_group,
    click_postpone_execution,
    click_prompt,
    get_scp_params,
    log_or_echo,
    pretty_print,
)

logger = logging.getLogger(__name__)


def handle_error(e: Exception, default_exception_msg, target="Credential password"):
    if isinstance(e, InvalidPinError):
        attempts = e.attempts_remaining
        if attempts:
            raise CliFail(f"Wrong {target}, {attempts} attempts remaining.")
        else:
            raise CliFail(f"{target} is blocked.")
    elif isinstance(e, ApduError):
        if e.sw == SW.AUTH_METHOD_BLOCKED:
            raise CliFail("A credential with the provided label already exists.")
        elif e.sw == SW.NO_SPACE:
            raise CliFail("No space left on the YubiKey for YubiHSM Auth credentials.")
        elif e.sw == SW.FILE_NOT_FOUND:
            raise CliFail("Credential with the provided label was not found.")
        elif e.sw == SW.SECURITY_CONDITION_NOT_SATISFIED:
            raise CliFail("The device was not touched.")
        elif e.sw == SW.CONDITIONS_NOT_SATISFIED:
            raise CliFail(f"{target} does not meet complexity requirement.")
    elif isinstance(e, ValueError):
        raise  # Re-raise, ValueErrors are handled elsewhere
    raise CliFail(default_exception_msg)


def _parse_touch_required(touch_required: bool) -> str:
    if touch_required:
        return "On"
    else:
        return "Off"


def _parse_algorithm(algorithm: ALGORITHM) -> str:
    if algorithm == ALGORITHM.AES128_YUBICO_AUTHENTICATION:
        return "Symmetric"
    else:
        return "Asymmetric"


def _parse_hex(hex):
    try:
        return bytes.fromhex(hex)
    except Exception as e:
        raise ValueError(e)


def _parse_key(key, key_len, key_type):
    key = _parse_hex(key)

    if len(key) != key_len:
        raise ValueError(
            f"must be exactly {key_len} bytes long ({key_len * 2} hexadecimal digits) "
            "long"
        )
    return key


def _parse_password(value, key_len, name):
    encoded = value.encode()
    if len(encoded) <= key_len:
        return encoded.ljust(key_len, b"\0")
    if len(encoded) == key_len * 2:
        return _parse_hex(value)
    raise ValueError(f"{name} must be at most 16 bytes")


@click_callback()
def click_parse_management_key(ctx, param, val):
    return _parse_key(val, MANAGEMENT_KEY_LEN, "Management key")


@click_callback()
def click_parse_management_password(ctx, param, val):
    return _parse_password(val, MANAGEMENT_KEY_LEN, "Management password")


@click_callback()
def click_parse_credential_password(ctx, param, val):
    return _parse_password(val, CREDENTIAL_PASSWORD_LEN, "Credential password")


@click_callback()
def click_parse_enc_key(ctx, param, val):
    return _parse_key(val, ALGORITHM.AES128_YUBICO_AUTHENTICATION.key_len, "ENC key")


@click_callback()
def click_parse_mac_key(ctx, param, val):
    return _parse_key(val, ALGORITHM.AES128_YUBICO_AUTHENTICATION.key_len, "MAC key")


@click_callback()
def click_parse_card_crypto(ctx, param, val):
    return _parse_hex(val)


@click_callback()
def click_parse_context(ctx, param, val):
    return _parse_hex(val)


def _prompt_management_key(prompt="Enter management password", confirm=False):
    management_password = click_prompt(
        prompt,
        default="",
        hide_input=True,
        show_default=False,
        confirmation_prompt=confirm,
    )
    return _parse_password(
        management_password, MANAGEMENT_KEY_LEN, "Management password"
    )


def _prompt_credential_password(prompt="Enter credential password", confirm=False):
    credential_password = click_prompt(
        prompt,
        hide_input=True,
        confirmation_prompt=confirm,
    )

    return _parse_password(
        credential_password, CREDENTIAL_PASSWORD_LEN, "Credential password"
    )


def _prompt_symmetric_key(name):
    symmetric_key = click_prompt(f"Enter {name}")

    return _parse_key(
        symmetric_key, ALGORITHM.AES128_YUBICO_AUTHENTICATION.key_len, name
    )


def _fname(fobj):
    return getattr(fobj, "name", fobj)


click_credential_password_option = click.option(
    "-c",
    "--credential-password",
    help="password to protect credential",
    callback=click_parse_credential_password,
)

click_management_key_option = click.option(
    "-m",
    "--management-password",
    "--management-key",
    "management_key",
    help="the management password",
    callback=click_parse_management_password,
)


click_touch_option = click.option(
    "-t", "--touch", is_flag=True, help="require touch on YubiKey to access credential"
)


@click_group(connections=[SmartCardConnection])
@click.pass_context
@click_postpone_execution
def hsmauth(ctx):
    """
    Manage the YubiHSM Auth application
    """
    dev = ctx.obj["device"]
    conn = dev.open_connection(SmartCardConnection)
    ctx.call_on_close(conn.close)

    scp_params = get_scp_params(ctx, CAPABILITY.HSMAUTH, conn)
    ctx.obj["session"] = HsmAuthSession(conn, scp_params)

    try:
        if not scp_params and dev.transport == TRANSPORT.NFC:
            # Dummy command to test access
            ctx.obj["session"].get_management_key_retries()
    except ApduError as e:
        if e.sw == SW.CONDITIONS_NOT_SATISFIED:
            raise CliFail("Unable to manage HSMAuth over NFC without SCP.")
        raise

    info = ctx.obj["info"]
    ctx.obj["fips_unready"] = (
        CAPABILITY.HSMAUTH in info.fips_capable
        and CAPABILITY.HSMAUTH not in info.fips_approved
    )


@hsmauth.command()
@click.pass_context
def info(ctx):
    """
    Display general status of the YubiHSM Auth application.
    """
    info = ctx.obj["info"]
    data = get_hsmauth_info(ctx.obj["session"])
    if CAPABILITY.HSMAUTH in info.fips_capable:
        # This is a bit ugly as it makes assumptions about the structure of data
        data["FIPS approved"] = CAPABILITY.HSMAUTH in info.fips_approved

    click.echo("\n".join(pretty_print(data)))


@hsmauth.command()
@click.pass_context
@click_force_option
def reset(ctx, force):
    """
    Reset all YubiHSM Auth data.

    This action will wipe all data and restore factory setting for
    the YubiHSM Auth application on the YubiKey.
    """

    if not force:
        click.confirm(
            "WARNING! This will delete all stored YubiHSM Auth data and restore "
            "factory setting. Proceed?",
            abort=True,
            err=True,
        )

    click.echo("Resetting YubiHSM Auth data...")
    ctx.obj["session"].reset()

    click.echo(
        "Reset complete. All YubiHSM Auth data has been cleared from the YubiKey."
    )
    click.echo("Your YubiKey now has an empty Management password.")


@hsmauth.group()
def credentials():
    """Manage YubiHSM Auth credentials."""


@credentials.command()
@click.pass_context
def list(ctx):
    """
    List all credentials.

    List all credentials stored on the YubiKey.
    """
    session = ctx.obj["session"]
    creds = session.list_credentials()

    if len(creds) == 0:
        click.echo("No items found")
    else:
        click.echo(f"Found {len(creds)} item(s)")

        max_size_label = max(len(cred.label) for cred in creds)
        max_size_type = (
            10
            if any(
                c.algorithm == ALGORITHM.EC_P256_YUBICO_AUTHENTICATION for c in creds
            )
            else 9
        )

        format_str = "{0: <{label_width}}\t{1: <{type_width}}\t{2}\t{3}"

        click.echo(
            format_str.format(
                "Label",
                "Type",
                "Touch",
                "Retries",
                label_width=max_size_label,
                type_width=max_size_type,
            )
        )

        for cred in creds:
            click.echo(
                format_str.format(
                    cred.label,
                    _parse_algorithm(cred.algorithm),
                    _parse_touch_required(cred.touch_required),
                    cred.counter,
                    label_width=max_size_label,
                    type_width=max_size_type,
                )
            )


@credentials.command()
@click.pass_context
@click.argument("label")
@click_credential_password_option
@click_management_key_option
@click_touch_option
def generate(ctx, label, credential_password, management_key, touch):
    """Generate an asymmetric credential.

    This will generate an asymmetric YubiHSM Auth credential
    (private key) on the YubiKey.

    \b
    LABEL label for the YubiHSM Auth credential
    """

    if ctx.obj["fips_unready"]:
        raise CliFail(
            "YubiKey FIPS must be in FIPS approved mode prior to adding credentials."
        )

    if management_key is None:
        management_key = _prompt_management_key()

    if credential_password is None:
        credential_password = _prompt_credential_password()

    session = ctx.obj["session"]

    try:
        session.generate_credential_asymmetric(
            management_key, label, credential_password, touch
        )
        click.echo("Asymmetric credential generated.")
    except Exception as e:
        handle_error(
            e, default_exception_msg="Failed to generate asymmetric credential."
        )


@credentials.command("import")
@click.pass_context
@click.argument("label")
@click.argument("private-key", type=click.File("rb"), metavar="PRIVATE-KEY")
@click.option("-p", "--password", help="password used to decrypt the private key")
@click_credential_password_option
@click_management_key_option
@click_touch_option
def import_credential(
    ctx, label, private_key, password, credential_password, management_key, touch
):
    """Import an asymmetric credential.

    This will import a private key as an asymmetric YubiHSM Auth credential
    to the YubiKey.

    \b
    LABEL        label for the YubiHSM Auth credential
    PRIVATE-KEY  file containing the private key (use '-' to use stdin)
    """
    if ctx.obj["fips_unready"]:
        raise CliFail(
            "YubiKey FIPS must be in FIPS approved mode prior to adding credentials."
        )

    if management_key is None:
        management_key = _prompt_management_key()

    if credential_password is None:
        credential_password = _prompt_credential_password()

    session = ctx.obj["session"]

    data = private_key.read()

    while True:
        if password is not None:
            password = password.encode()
        try:
            private_key = parse_private_key(data, password)
        except InvalidPasswordError:
            logger.debug("Error parsing key", exc_info=True)
            if password is None:
                password = click_prompt(
                    "Enter password to decrypt key",
                    default="",
                    hide_input=True,
                    show_default=False,
                )
                continue
            else:
                password = None
                click.echo("Wrong password.")
            continue
        break

    try:
        session.put_credential_asymmetric(
            management_key,
            label,
            private_key,
            credential_password,
            touch,
        )
        click.echo("Asymmetric credential imported.")
    except Exception as e:
        handle_error(e, default_exception_msg="Failed to import asymmetric credential.")


@credentials.command()
@click.pass_context
@click.argument("label")
@click.argument("public-key-output", type=click.File("wb"), metavar="PUBLIC-KEY")
@click_format_option
def export(ctx, label, public_key_output, format):
    """Export the public key corresponding to an asymmetric credential.

    This will export the long-term public key corresponding to the
    asymmetric YubiHSM Auth credential stored on the YubiKey.

    \b
    LABEL      label for the YubiHSM Auth credential
    PUBLIC-KEY file to write the public key to (use '-' to use stdout)
    """

    session = ctx.obj["session"]

    try:
        public_key = session.get_public_key(label)
        key_encoding = format
        public_key_encoded = public_key.public_bytes(
            encoding=key_encoding,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

        public_key_output.write(public_key_encoded)

        log_or_echo(
            f"Public key for {label} written to {_fname(public_key_output)}",
            logger,
            public_key_output,
        )
    except ApduError as e:
        if e.sw == SW.AUTH_METHOD_BLOCKED:
            raise CliFail("The entry is not an asymmetric credential.")
        elif e.sw == SW.FILE_NOT_FOUND:
            raise CliFail("Credential not found.")
        else:
            raise CliFail("Unable to export public key.")


@credentials.command()
@click.pass_context
@click.argument("label")
@click.option("-E", "--enc-key", help="the ENC key", callback=click_parse_enc_key)
@click.option("-M", "--mac-key", help="the MAC key", callback=click_parse_mac_key)
@click.option(
    "-g", "--generate", is_flag=True, help="generate a random encryption and mac key"
)
@click_credential_password_option
@click_management_key_option
@click_touch_option
def symmetric(
    ctx, label, credential_password, management_key, enc_key, mac_key, generate, touch
):
    """Import a symmetric credential.

    This will import an encryption and mac key as a symmetric YubiHSM Auth credential on
    the YubiKey.

    \b
    LABEL  label for the YubiHSM Auth credential
    """

    if ctx.obj["fips_unready"]:
        raise CliFail(
            "YubiKey FIPS must be in FIPS approved mode prior to adding credentials."
        )

    if management_key is None:
        management_key = _prompt_management_key()

    if credential_password is None:
        credential_password = _prompt_credential_password()

    if generate and (enc_key or mac_key):
        ctx.fail("--enc-key and --mac-key cannot be combined with --generate")

    if generate:
        enc_key = os.urandom(ALGORITHM.AES128_YUBICO_AUTHENTICATION.key_len)
        mac_key = os.urandom(ALGORITHM.AES128_YUBICO_AUTHENTICATION.key_len)
        click.echo("Generated ENC and MAC keys:")
        click.echo("\n".join(pretty_print({"ENC-KEY": enc_key, "MAC-KEY": mac_key})))

    if not enc_key:
        enc_key = _prompt_symmetric_key("ENC key")

    if not mac_key:
        mac_key = _prompt_symmetric_key("MAC key")

    session = ctx.obj["session"]

    try:
        session.put_credential_symmetric(
            management_key,
            label,
            enc_key,
            mac_key,
            credential_password,
            touch,
        )
        click.echo("Symmetric credential stored.")

    except Exception as e:
        handle_error(e, default_exception_msg="Failed to import symmetric credential.")


@credentials.command()
@click.pass_context
@click.argument("label")
@click.option(
    "-d", "--derivation-password", help="deriviation password for ENC and MAC keys"
)
@click_credential_password_option
@click_management_key_option
@click_touch_option
def derive(ctx, label, derivation_password, credential_password, management_key, touch):
    """Import a symmetric credential derived from a password.

    This will import a symmetric YubiHSM Auth credential by deriving
    ENC and MAC keys from a password.

    \b
    LABEL  label for the YubiHSM Auth credential
    """

    if ctx.obj["fips_unready"]:
        raise CliFail(
            "YubiKey FIPS must be in FIPS approved mode prior to adding credentials."
        )

    if management_key is None:
        management_key = _prompt_management_key()

    if credential_password is None:
        credential_password = _prompt_credential_password()

    if derivation_password is None:
        derivation_password = click_prompt(
            "Enter derivation password",
            hide_input=True,
            confirmation_prompt=True,
        )

    session = ctx.obj["session"]

    try:
        session.put_credential_derived(
            management_key, label, derivation_password, credential_password, touch
        )
        click.echo("Derived symmetric credential stored.")
    except Exception as e:
        handle_error(e, default_exception_msg="Failed to import symmetric credential.")


@credentials.command()
@click.pass_context
@click.argument("label")
@click_management_key_option
@click_force_option
def delete(ctx, label, management_key, force):
    """
    Delete a credential.

    This will delete a YubiHSM Auth credential from the YubiKey.

    \b
    LABEL a label to match a single credential (as shown in "list")
    """

    if management_key is None:
        management_key = _prompt_management_key()

    if not force:
        click.confirm(
            f"Delete credential: {label} ?",
            abort=True,
            err=True,
        )

    session = ctx.obj["session"]

    try:
        session.delete_credential(management_key, label)
        click.echo("Credential deleted.")
    except Exception as e:
        handle_error(
            e,
            default_exception_msg="Failed to delete credential.",
        )


@credentials.command()
@click.pass_context
@click.argument("label")
@click.option(
    "-c",
    "--credential-password",
    help="current credential password",
    callback=click_parse_credential_password,
)
@click_management_key_option
@click.option(
    "-n",
    "--new-credential-password",
    help="a new credential password to set",
    callback=click_parse_credential_password,
)
def change_password(
    ctx, label, credential_password, management_key, new_credential_password
):
    """
    Change the password of a credential.

    This will change the password of a YubiHSM Auth credential stored on the YubiKey.

    Changing a password requires either the current password for the credential,
    or the management password (not both).

    \b
    LABEL a label to match a single credential (as shown in "list")
    """

    session = ctx.obj["session"]
    if credential_password and management_key:
        ctx.fail("--credential-password and --management-password can't be combined")

    if management_key is None and credential_password is None:
        click.echo("Choose authentication method:")
        click.echo(" [1] Current credential password")
        click.echo(" [2] Management password")
        method = click_prompt(
            "Enter selection",
            type=click.Choice([1, 2]),
            default=1,
            show_choices=False,
        )
        if method == 1:
            credential_password = _prompt_credential_password(
                "Enter current credential password"
            )
        else:
            management_key = _prompt_management_key()

    use_management_key = management_key is not None
    if new_credential_password is None:
        new_credential_password = _prompt_credential_password(
            "Enter the new credential password", confirm=True
        )

    try:
        if use_management_key:
            session.change_credential_password_admin(
                label, management_key, new_credential_password
            )
        else:
            session.change_credential_password(
                label, credential_password, new_credential_password
            )
        click.echo("Credential password changed.")
    except Exception as e:
        handle_error(
            e,
            default_exception_msg="Failed to change credential password.",
            target="Management password"
            if use_management_key
            else "Credential password",
        )


@hsmauth.group()
def access():
    """Manage Management Key for YubiHSM Auth"""


@access.command(hidden=True)
@click.pass_context
@click.option(
    "-m",
    "--management-key",
    help="current management key",
    show_default=True,
    callback=click_parse_management_key,
)
@click.option(
    "-n",
    "--new-management-key",
    help="a new management key to set",
    callback=click_parse_management_key,
)
@click.option(
    "-g",
    "--generate",
    is_flag=True,
    help="generate a random management key (can't be used with --new-management-key)",
)
def change_management_key(ctx, management_key, new_management_key, generate):
    """
    Change the management key.

    Allows you to change the management key which is required to add and delete
    YubiHSM Auth credentials stored on the YubiKey.
    """

    if management_key is None:
        management_key = _prompt_management_key(
            "Enter current management key [blank to use default key]"
        )

    session = ctx.obj["session"]

    # Can't combine new key with generate.
    if new_management_key and generate:
        ctx.fail("Invalid options: --new-management-key conflicts with --generate")

    if new_management_key is None:
        if generate:
            if ctx.obj["info"].version >= (5, 8):
                raise CliFail(
                    "Generating a random management key is not supported on this "
                    "YubiKey."
                )
            new_management_key = generate_random_management_key()
            click.echo(f"Generated management key: {new_management_key.hex()}")
        else:
            try:
                new_management_key = bytes.fromhex(
                    click_prompt(
                        "Enter the new management key",
                        hide_input=True,
                        confirmation_prompt=True,
                    )
                )
            except Exception:
                ctx.fail("New management key has the wrong format.")

    if new_management_key is None:
        raise CliFail("No new management key provided.")

    if len(new_management_key) != MANAGEMENT_KEY_LEN:
        raise CliFail(
            "Management key has the wrong length (expected %d bytes)."
            % MANAGEMENT_KEY_LEN
        )

    try:
        session.put_management_key(management_key, new_management_key)
        click.echo("Management key changed.")
    except Exception as e:
        handle_error(
            e,
            default_exception_msg="Failed to change management key.",
            target="Management key",
        )


@access.command()
@click.pass_context
@click.option(
    "-m",
    "--management-password",
    "management_key",
    help="current management password",
    show_default=True,
    callback=click_parse_management_password,
)
@click.option(
    "-n",
    "--new-management-password",
    "new_management_key",
    help="a new management password to set",
    callback=click_parse_management_password,
)
def change_management_password(ctx, management_key, new_management_key):
    """
    Change the management key.

    Allows you to change the management key which is required to add and delete
    YubiHSM Auth credentials stored on the YubiKey.
    """

    if management_key is None:
        management_key = _prompt_management_key(
            "Enter your current management password",
        )

    if new_management_key is None:
        new_management_key = _prompt_management_key(
            "Enter a new management password", confirm=True
        )

    session = ctx.obj["session"]
    try:
        session.put_management_key(management_key, new_management_key)
        click.echo("Management password changed.")
    except Exception as e:
        handle_error(
            e,
            default_exception_msg="Failed to change management password.",
            target="Management password",
        )



# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import re

import click

from yubikit.core import TRANSPORT, YUBIKEY
from yubikit.core.fido import FidoConnection
from yubikit.core.otp import OtpConnection
from yubikit.core.smartcard import SmartCardConnection
from yubikit.management import (
    CAPABILITY,
    DEVICE_FLAG,
    USB_INTERFACE,
    DeviceConfig,
    ManagementSession,
    Mode,
)

from .util import (
    CliFail,
    EnumChoice,
    click_force_option,
    click_group,
    click_postpone_execution,
    click_prompt,
)

logger = logging.getLogger(__name__)


CLEAR_LOCK_CODE = b"\0" * 16


def prompt_lock_code():
    return click_prompt("Enter your lock code", hide_input=True)


@click_group(connections=[SmartCardConnection, OtpConnection, FidoConnection])
@click.pass_context
@click_postpone_execution
def config(ctx):
    """
    Configure the YubiKey, enable or disable applications.

    The applications may be enabled and disabled independently
    over different transports (USB and NFC). The configuration may
    also be protected by a lock code.

    Examples:

    \b
      Disable PIV over NFC:
      $ ykman config nfc --disable PIV

    \b
      Enable all applications over USB:
      $ ykman config usb --enable-all

    \b
      Generate and set a random application lock code:
      $ ykman config set-lock-code --generate
    """
    dev = ctx.obj["device"]
    for conn_type in (SmartCardConnection, OtpConnection, FidoConnection):
        if dev.supports_connection(conn_type):
            try:
                conn = dev.open_connection(conn_type)
                ctx.call_on_close(conn.close)
                ctx.obj["session"] = ManagementSession(conn)
                return
            except Exception:
                logger.warning(
                    f"Failed connecting to the YubiKey over {conn_type}", exc_info=True
                )
    raise CliFail("Couldn't connect to the YubiKey.")


def _require_config(ctx):
    info = ctx.obj["info"]
    if (1, 0, 0) < info.version < (5, 0, 0):
        raise CliFail(
            "Configuring applications is not supported on this YubiKey. "
            "Use the `mode` command to configure USB interfaces."
        )


@config.command()
@click.pass_context
@click_force_option
def reset(ctx, force):
    """
    Reset all YubiKey data.

    This command is used with the YubiKey Bio Multi-protocol Edition.

    This action will wipe all data and restore factory settings for
    all applications on the YubiKey.
    """
    dev = ctx.obj["device"]
    if not dev.supports_connection(SmartCardConnection):
        raise CliFail("Full device reset requires the CCID interface to be enabled.")

    info = ctx.obj["info"]
    # reset_blocked is a sure indicator of the command
    if not info.reset_blocked:
        # No reset blocked, we can still check for Bio MPE
        transport = ctx.obj["device"].transport
        has_piv = CAPABILITY.PIV in info.supported_capabilities.get(transport)
        if not (info._is_bio and has_piv):
            raise CliFail(
                "Full device reset is not supported on this YubiKey, "
                "refer to reset commands for specific applications instead."
            )

    if not force:
        click.confirm(
            "WARNING! This will delete all stored data and restore factory "
            "settings. Proceed?",
            abort=True,
            err=True,
        )

    click.echo("Resetting YubiKey data...")
    ctx.obj["session"].device_reset()

    click.echo("Reset complete. All data has been cleared from the YubiKey.")


@config.command("set-lock-code")
@click.pass_context
@click_force_option
@click.option("-l", "--lock-code", metavar="HEX", help="current lock code")
@click.option(
    "-n",
    "--new-lock-code",
    metavar="HEX",
    help="new lock code (can't be used with --generate)",
)
@click.option("-c", "--clear", is_flag=True, help="clear the lock code")
@click.option(
    "-g",
    "--generate",
    is_flag=True,
    help="generate a random lock code (can't be used with --new-lock-code)",
)
def set_lock_code(ctx, lock_code, new_lock_code, clear, generate, force):
    """
    Set or change the configuration lock code.

    A lock code may be used to protect the application configuration.
    The lock code must be a 32 characters (16 bytes) hex value.
    """

    _require_config(ctx)
    info = ctx.obj["info"]
    app = ctx.obj["session"]

    if sum(1 for arg in [new_lock_code, generate, clear] if arg) > 1:
        raise CliFail(
            "Invalid options: Only one of --new-lock-code, --generate, "
            "and --clear may be used."
        )

    # Get the new lock code to set
    if clear:
        set_code = CLEAR_LOCK_CODE
    elif generate:
        set_code = os.urandom(16)
        click.echo(f"Using a randomly generated lock code: {set_code.hex()}")
        if not force:
            click.confirm(
                "Lock configuration with this lock code?", abort=True, err=True
            )
    else:
        if not new_lock_code:
            new_lock_code = click_prompt(
                "Enter your new lock code", hide_input=True, confirmation_prompt=True
            )
        set_code = _parse_lock_code(new_lock_code)

    # Get the current lock code to use
    if info.is_locked:
        if not lock_code:
            lock_code = click_prompt("Enter your current lock code", hide_input=True)
        use_code = _parse_lock_code(lock_code)
    else:
        if lock_code:
            raise CliFail(
                "No lock code is currently set. Use --new-lock-code to set one."
            )
        use_code = None

    # Set new lock code
    try:
        app.write_device_config(
            None,
            False,
            use_code,
            set_code,
        )
        click.echo("Lock code updated.")
    except Exception:
        if info.is_locked:
            raise CliFail("Failed to change the lock code. Wrong current code?")
        raise CliFail("Failed to set the lock code.")


def _get_lock_code(is_locked, lock_code, force):
    if force and is_locked and not lock_code:
        raise CliFail("Configuration is locked - supply the --lock-code option.")
    if lock_code and not is_locked:
        raise CliFail("Configuration is not locked - remove the --lock-code option.")

    if is_locked and not lock_code:
        lock_code = prompt_lock_code()

    if lock_code:
        lock_code = _parse_lock_code(lock_code)

    return lock_code


def _configure_applications(
    ctx,
    config,
    changes,
    transport,
    enable,
    disable,
    lock_code,
    force,
):
    info = ctx.obj["info"]

    # If any app reset is blocked, we will not be able to toggle applications
    if info.reset_blocked:
        raise CliFail(
            "This YubiKey must be in a newly reset state before applications can be "
            "toggled."
        )

    supported = info.supported_capabilities.get(transport)
    enabled = info.config.enabled_capabilities.get(transport)

    if not supported:
        raise CliFail(f"{transport} not supported on this YubiKey.")

    if enable & disable:
        raise CliFail("Invalid options.")

    unsupported = ~supported & (enable | disable)
    if unsupported:
        raise CliFail(
            f"{unsupported.display_name} not supported over {transport} on this "
            "YubiKey."
        )

    # N.B. NOT (~) of IntFlag doesn't work as expected
    new_enabled = (enabled | enable) & ~int(disable)

    if transport == TRANSPORT.USB:
        if sum(CAPABILITY) & new_enabled == 0:
            raise CliFail(f"Can not disable all applications over {transport}.")

        reboot = enabled.usb_interfaces != new_enabled.usb_interfaces
    else:
        reboot = False

    if reboot:
        changes.append("The YubiKey will reboot")

    if lock_code and not info.is_locked:
        raise CliFail("Configuration is not locked - remove the --lock-code option.")

    click.echo(f"{transport} configuration changes:")
    for change in changes:
        click.echo(f"  {change}")
    if not force:
        click.confirm("Proceed?", abort=True, err=True)

    config.enabled_capabilities = {transport: new_enabled}

    lock_code = _get_lock_code(info.is_locked, lock_code, force)

    app = ctx.obj["session"]
    try:
        app.write_device_config(
            config,
            reboot,
            lock_code,
        )
        click.echo(f"{transport} application configuration updated.")
    except Exception:
        raise CliFail(f"Failed to configure {transport} applications.")


@config.command()
@click.pass_context
@click_force_option
@click.option(
    "-e",
    "--enable",
    multiple=True,
    type=EnumChoice(CAPABILITY),
    help="enable applications",
)
@click.option(
    "-d",
    "--disable",
    multiple=True,
    type=EnumChoice(CAPABILITY),
    help="disable applications",
)
@click.option(
    "-l", "--list", "list_enabled", is_flag=True, help="list enabled applications"
)
@click.option("-a", "--enable-all", is_flag=True, help="enable all applications")
@click.option(
    "-L",
    "--lock-code",
    metavar="HEX",
    help="current application configuration lock code",
)
@click.option(
    "--touch-eject",
    is_flag=True,
    help="when set, the button toggles the state"
    " of the smartcard between ejected and inserted (CCID only)",
)
@click.option("--no-touch-eject", is_flag=True, help="disable touch eject (CCID only)")
@click.option(
    "--autoeject-timeout",
    required=False,
    type=int,
    default=None,
    metavar="SECONDS",
    help="when set, the smartcard will automatically eject"
    " after the given time (implies --touch-eject)",
)
@click.option(
    "--chalresp-timeout",
    required=False,
    type=int,
    default=None,
    metavar="SECONDS",
    help="sets the timeout when waiting for touch for challenge-response in the OTP "
    "application",
)
def usb(
    ctx,
    enable,
    disable,
    list_enabled,
    enable_all,
    touch_eject,
    no_touch_eject,
    autoeject_timeout,
    chalresp_timeout,
    lock_code,
    force,
):
    """
    Enable or disable applications over USB.
    """
    _require_config(ctx)

    if not any(
        [
            list_enabled,
            enable_all,
            enable,
            disable,
            touch_eject,
            no_touch_eject,
            autoeject_timeout,
            chalresp_timeout,
        ]
    ):
        raise CliFail("No configuration options chosen.")

    if touch_eject and no_touch_eject:
        raise CliFail("Invalid options.")

    if list_enabled:
        _list_apps(ctx, TRANSPORT.USB)

    config = DeviceConfig({}, autoeject_timeout, chalresp_timeout)
    changes = []
    info = ctx.obj["info"]

    if enable_all:
        enable = info.supported_capabilities.get(TRANSPORT.USB)
        changes.append("Enable all applications")
    else:
        enable = CAPABILITY(sum(enable))
        if enable:
            changes.append(f"Enable {enable.display_name}")
    disable = CAPABILITY(sum(disable))
    if disable:
        changes.append(f"Disable {disable.display_name}")

    if touch_eject:
        config.device_flags = info.config.device_flags | DEVICE_FLAG.EJECT
        changes.append("Enable touch-eject")
    if no_touch_eject:
        config.device_flags = info.config.device_flags & ~DEVICE_FLAG.EJECT
        changes.append("Disable touch-eject")
    if autoeject_timeout:
        changes.append(f"Set auto-eject timeout to {autoeject_timeout}")
    if chalresp_timeout:
        changes.append(f"Set challenge-response timeout to {chalresp_timeout}")

    _configure_applications(
        ctx,
        config,
        changes,
        TRANSPORT.USB,
        enable,
        disable,
        lock_code,
        force,
    )


@config.command()
@click.pass_context
@click_force_option
@click.option(
    "-e",
    "--enable",
    multiple=True,
    type=EnumChoice(CAPABILITY),
    help="enable applications",
)
@click.option(
    "-d",
    "--disable",
    multiple=True,
    type=EnumChoice(CAPABILITY),
    help="disable applications",
)
@click.option("-a", "--enable-all", is_flag=True, help="enable all applications")
@click.option("-D", "--disable-all", is_flag=True, help="disable all applications")
@click.option(
    "-l", "--list", "list_enabled", is_flag=True, help="list enabled applications"
)
@click.option(
    "-L",
    "--lock-code",
    metavar="HEX",
    help="current application configuration lock code",
)
@click.option(
    "-R",
    "--restrict",
    is_flag=True,
    help="Disable NFC for transport, re-enabled by USB power",
)
def nfc(
    ctx,
    enable,
    disable,
    enable_all,
    disable_all,
    list_enabled,
    lock_code,
    restrict,
    force,
):
    """
    Enable or disable applications over NFC.
    """
    info = ctx.obj["info"]
    if TRANSPORT.NFC not in info.supported_capabilities:
        raise CliFail("This YubiKey does not support NFC.")

    _require_config(ctx)

    if not any([list_enabled, enable_all, enable, disable_all, disable, restrict]):
        raise CliFail("No configuration options chosen.")

    if list_enabled:
        _list_apps(ctx, TRANSPORT.NFC)

    config = DeviceConfig({}, None, None, None)

    if restrict:
        if info.version < (5, 7):
            raise CliFail("NFC restriction requires YubiKey 5.7 or later.")

        config.nfc_restricted = True
        lock_code = _get_lock_code(info.is_locked, lock_code, force)
        ctx.obj["session"].write_device_config(config, False, lock_code)
        click.echo(
            "YubiKey NFC disabled. It will be re-enabled automatically the next time "
            "it is connected to USB power."
        )
        ctx.exit()

    changes = []

    nfc_supported = info.supported_capabilities.get(TRANSPORT.NFC)
    if enable_all:
        enable = nfc_supported
        changes.append("Enable all applications")
    else:
        enable = CAPABILITY(sum(enable))
        if enable:
            changes.append(f"Enable {enable.display_name}")

    if disable_all:
        disable = nfc_supported
        changes.append("Disable all applications")
    else:
        disable = CAPABILITY(sum(disable))
        if disable:
            changes.append(f"Disable {disable.display_name}")

    _configure_applications(
        ctx,
        config,
        changes,
        TRANSPORT.NFC,
        enable,
        disable,
        lock_code,
        force,
    )


def _list_apps(ctx, transport):
    enabled = ctx.obj["info"].config.enabled_capabilities.get(transport)
    if enabled is None:
        raise CliFail(f"{transport} not supported on this YubiKey.")

    for app in CAPABILITY:
        if app & enabled:
            click.echo(app.display_name)
    ctx.exit()


def _ensure_not_invalid_options(ctx, enable, disable):
    if enable & disable:
        raise CliFail("Invalid options.")


def _parse_lock_code(lock_code):
    try:
        lock_code = bytes.fromhex(lock_code)
        if lock_code and len(lock_code) != 16:
            raise CliFail(
                "Lock code must be exactly 16 bytes (32 hexadecimal digits) long."
            )
        return lock_code
    except Exception:
        raise CliFail("Lock code has the wrong format.")


# MODE


def _parse_interface_string(interface):
    for iface in USB_INTERFACE:
        if (iface.name or "").startswith(interface):
            return iface
    raise ValueError()


def _parse_mode_string(ctx, param, mode):
    try:
        mode_int = int(mode)
        try:
            return Mode.from_code(mode_int)
        except IndexError:
            raise CliFail(f"Invalid mode: {mode_int}")
    except ValueError:
        pass  # Not a numeric mode, parse string

    try:
        if mode[0] in ["+", "-"]:
            info = ctx.obj["info"]
            usb_enabled = info.config.enabled_capabilities[TRANSPORT.USB]
            interfaces = usb_enabled.usb_interfaces
            for mod in re.findall(r"[+-][A-Z]+", mode.upper()):
                interface = _parse_interface_string(mod[1:])
                if mod.startswith("+"):
                    interfaces |= interface
                else:
                    interfaces ^= interface
        else:
            interfaces = USB_INTERFACE(0)
            for t in re.split(r"[+]+", mode.upper()):
                if t:
                    interfaces |= _parse_interface_string(t)
    except ValueError:
        raise CliFail(f"Invalid mode string: {mode}")

    return Mode(interfaces)


@config.command()
@click.argument("mode", callback=_parse_mode_string)
@click.option(
    "--touch-eject",
    is_flag=True,
    help="when set, the button "
    "toggles the state of the smartcard between ejected and inserted "
    "(CCID only)",
)
@click.option(
    "--autoeject-timeout",
    required=False,
    type=int,
    default=0,
    metavar="SECONDS",
    help="when set, the smartcard will automatically eject after the given time "
    "(implies --touch-eject, CCID only)",
)
@click.option(
    "--chalresp-timeout",
    required=False,
    type=int,
    default=0,
    metavar="SECONDS",
    help="sets the timeout when waiting for touch for challenge response",
)
@click_force_option
@click.pass_context
def mode(ctx, mode, touch_eject, autoeject_timeout, chalresp_timeout, force):
    """
    Manage connection modes (USB Interfaces).

    This command is generally used with YubiKeys prior to the 5 series.
    Use "ykman config usb" for more granular control on YubiKey 5 and later.

    Get the current connection mode of the YubiKey, or set it to MODE.

    MODE can be a string, such as "OTP+FIDO+CCID", or a shortened form: "o+f+c".
    It can also be a mode number.

    Examples:

    \b
      Set the OTP and FIDO mode:
      $ ykman config mode OTP+FIDO

    \b
      Set the CCID only mode and use touch to eject the smart card:
      $ ykman config mode CCID --touch-eject
    """
    info = ctx.obj["info"]
    mgmt = ctx.obj["session"]
    usb_enabled = info.config.enabled_capabilities[TRANSPORT.USB]
    my_mode = Mode(usb_enabled.usb_interfaces)
    usb_supported = info.supported_capabilities[TRANSPORT.USB]
    interfaces_supported = usb_supported.usb_interfaces
    pid = ctx.obj["pid"]
    if pid:
        key_type = pid.yubikey_type
    else:
        key_type = None

    if autoeject_timeout:  # autoeject implies touch eject
        touch_eject = True
    autoeject = autoeject_timeout if touch_eject else None

    if mode.interfaces != USB_INTERFACE.CCID:
        if touch_eject:
            raise CliFail("--touch-eject can only be used when setting CCID-only mode")

    if not force:
        if mode == my_mode:
            raise CliFail(f"Mode is already {mode}, nothing to do...", 0)
        elif key_type in (YUBIKEY.YKS, YUBIKEY.YKP):
            raise CliFail(
                "Mode switching is not supported on this YubiKey!\n"
                "Use --force to attempt to set it anyway."
            )
        elif mode.interfaces not in interfaces_supported:
            raise CliFail(
                f"Mode {mode} is not supported on this YubiKey!\n"
                + "Use --force to attempt to set it anyway."
            )
        elif info.is_sky and USB_INTERFACE.FIDO not in mode.interfaces:
            raise CliFail("Security Key requires FIDO to be enabled.")
        if not force:
            click.confirm(f"Set mode of YubiKey to {mode}?", abort=True, err=True)

    try:
        mgmt.set_mode(mode, chalresp_timeout, autoeject)
        logger.info("USB mode updated")
        click.echo(
            "Mode set! You must remove and re-insert your YubiKey "
            "for this change to take effect."
        )
    except Exception:
        raise CliFail(
            "Failed to switch mode on the YubiKey. Make sure your "
            "YubiKey does not have an access code set."
        )



# Copyright (c) 2015 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
from datetime import datetime, timezone
from typing import Any

from yubikit.core.smartcard import (
    AID,
    SW,
    ApduError,
    SmartCardConnection,
    SmartCardProtocol,
)
from yubikit.openpgp import (
    _INVALID_PIN,
    INS,
    KEY_REF,
    PW,
    AlgorithmAttributes,
    EcAttributes,
    KdfNone,
    OpenPgpSession,
    RsaAttributes,
)

logger = logging.getLogger(__name__)


def safe_reset(connection: SmartCardConnection) -> None:
    """Performs an OpenPGP factory reset while avoiding any unneccessary commands.

    If any data is unreadable preventing the OpenPgpSession from initializing, then
    OpenPgpSession.reset() will not be able to be called. This function can instead
    be  used to reset the application into a fresh state.
    """
    logger.debug("Attempting safe OpenPGP factory reset")
    protocol = SmartCardProtocol(connection)
    protocol.select(AID.OPENPGP)

    for pw in (PW.USER, PW.ADMIN):
        logger.debug(f"Verify {pw.name} PIN with invalid attempts until blocked")
        while True:
            try:
                protocol.send_apdu(0, INS.VERIFY, 0, pw, _INVALID_PIN)
            except ApduError as e:
                if e.sw == SW.SECURITY_CONDITION_NOT_SATISFIED:
                    continue
                # Either blocked, or an unexpected error, move to the next step
                break

    # Reset the application
    logger.debug("Sending TERMINATE, then ACTIVATE")
    protocol.send_apdu(0, INS.TERMINATE, 0, 0)
    protocol.send_apdu(0, INS.ACTIVATE, 0, 0)
    logger.info("OpenPGP application data reset performed")


def _format_ref(ref: KEY_REF) -> str:
    if ref == KEY_REF.SIG:
        return "Signature key"
    if ref == KEY_REF.DEC:
        return "Decryption key"
    if ref == KEY_REF.AUT:
        return "Authentication key"
    if ref == KEY_REF.ATT:
        return "Attestation key"
    return ref.name


def _format_fingerprint(fp: bytes) -> str:
    return "  ".join(
        " ".join(fp[h * 10 + s * 2 :][:2].hex() for s in range(5)) for h in range(2)
    ).upper()


def _format_date(timestamp: int) -> str:
    return datetime.fromtimestamp(timestamp, timezone.utc).isoformat()


def _format_algorithm(alg: AlgorithmAttributes) -> str:
    if isinstance(alg, RsaAttributes):
        return f"RSA{alg.n_len}"
    if isinstance(alg, EcAttributes):
        return f"{alg.oid}"
    return "Unknown key type"


def get_key_info(discretionary, ref, status):
    alg = discretionary.get_algorithm_attributes(ref)
    return {
        "Key slot": _format_ref(ref),
        "Fingerprint": _format_fingerprint(discretionary.fingerprints[ref]),
        "Algorithm": _format_algorithm(alg),
        "Origin": status.name if status is not None else "UNKNOWN",
        "Created": _format_date(discretionary.generation_times[ref]),
        "Touch policy": discretionary.get_uif(ref),
    }


def get_openpgp_info(session: OpenPgpSession):
    """Get human readable information about the OpenPGP configuration.

    :param session: The OpenPGP session.
    """
    data = session.get_application_related_data()
    discretionary = data.discretionary
    retries = discretionary.pw_status
    info: dict[str, Any] = {
        "OpenPGP version": "%d.%d" % data.aid.version,
        "Application version": "%d.%d.%d" % session.version,
        "PIN tries remaining": retries.attempts_user,
        "Reset code tries remaining": retries.attempts_reset,
        "Admin PIN tries remaining": retries.attempts_admin,
        "Require PIN for signature": retries.pin_policy_user,
        "KDF enabled": not isinstance(session.get_kdf(), KdfNone),
    }

    for ref, fp in discretionary.fingerprints.items():
        if session.version >= (5, 2, 0):
            if not discretionary.key_information[ref] or ref == KEY_REF.ATT:
                continue
        else:
            if not any(fp):
                continue

        info[_format_ref(ref)] = {
            "Fingerprint": _format_fingerprint(fp),
            "Touch policy": discretionary.get_uif(ref),
        }

    return info



# -*- coding: utf-8 -*-
import os
import re
import struct
import tempfile

import pytest
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from yubikit.core import TRANSPORT, Tlv
from yubikit.hsmauth import (
    INS_CALCULATE,
    TAG_CONTEXT,
    TAG_CREDENTIAL_PASSWORD,
    TAG_LABEL,
    _parse_credential_password,
    _parse_label,
)
from yubikit.management import CAPABILITY

from .. import condition

DEFAULT_MANAGEMENT_KEY = "00000000000000000000000000000000"
NON_DEFAULT_MANAGEMENT_KEY = "11111111111111111111111111111112"


# Test both password and key
@pytest.fixture(params=[DEFAULT_MANAGEMENT_KEY, "p4ssw0rd123"])
def management_key(request, ykman_cli, info):
    key = request.param
    if key == DEFAULT_MANAGEMENT_KEY and CAPABILITY.HSMAUTH in info.fips_capable:
        key = "00000000000000000000000000000001"

    if key != DEFAULT_MANAGEMENT_KEY:
        ykman_cli(
            "hsmauth",
            "access",
            "change-management-password",
            "-m",
            "",
            "-n",
            key,
        )

    yield key


def generate_pem_eccp256_keypair():
    pk = ec.generate_private_key(ec.SECP256R1(), default_backend())
    return (
        pk.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        pk.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ),
    )


@pytest.fixture()
def eccp256_keypair():
    tmp = tempfile.NamedTemporaryFile(delete=False)
    private_key, public_key = generate_pem_eccp256_keypair()
    tmp.write(private_key)
    tmp.close()
    yield tmp.name, public_key
    os.remove(tmp.name)


@pytest.fixture()
def tmp_file():
    tmp = tempfile.NamedTemporaryFile(delete=False)
    yield tmp
    tmp.close()
    os.remove(tmp.name)


@pytest.fixture(autouse=True)
@condition.capability(CAPABILITY.HSMAUTH)
@condition.min_version(5, 4, 3)
def preconditions(ykman_cli):
    ykman_cli("hsmauth", "reset", "-f")


class TestHsmAuth:
    def test_hsmauth_info(self, ykman_cli):
        output = ykman_cli("hsmauth", "info").output
        assert "version:" in output

    def test_hsmauth_reset(self, ykman_cli):
        output = ykman_cli("hsmauth", "reset", "-f").output
        assert "Reset complete." in output


def calculate_session_keys_apdu(label, context, credential_password):
    data = (
        Tlv(TAG_LABEL, _parse_label(label))
        + Tlv(TAG_CONTEXT, context)
        + Tlv(TAG_CREDENTIAL_PASSWORD, _parse_credential_password(credential_password))
    )

    apdu = struct.pack("<BBBB", 0, INS_CALCULATE, 0, 0).hex()
    apdu = apdu + ":" + data.hex() + "=9000"

    return apdu


class TestCredentials:
    def verify_credential_password(
        self, ykman_cli, transport, info, credential_password, label
    ):
        context = b"g\xfc\xf1\xfe\xb5\xf1\xd8\x83\xedv=\xbfI0\x90\xbb"
        apdu = calculate_session_keys_apdu(label, context, credential_password)

        # Try to calculate session keys using credential password
        # TODO: Use SCP if needed
        if transport == TRANSPORT.NFC and CAPABILITY.HSMAUTH in info.fips_capable:
            args = ("--scp-sd", "scp11b", "0")
        else:
            args = tuple()

        ykman_cli(*args, "apdu", "-a", "hsmauth", apdu)

    def import_key_derived(
        self,
        ykman_cli,
        management_key,
        label,
        credential_password="12345679",
        derivation_password="p4ssw0rd",
        touch_required=False,
    ):
        args = (
            "hsmauth",
            "credentials",
            "derive",
            label,
            "-c",
            credential_password,
            "-d",
            derivation_password,
            "-m",
            management_key,
        )
        if touch_required:
            args += ("-t",)
        ykman_cli(*args)

    def test_import_credential_symmetric(
        self, ykman_cli, transport, info, management_key
    ):
        ykman_cli(
            "hsmauth",
            "credentials",
            "symmetric",
            "test-name-sym",
            "-c",
            "12345679",
            "-E",
            os.urandom(16).hex(),
            "-M",
            os.urandom(16).hex(),
            "-m",
            management_key,
        )
        self.verify_credential_password(
            ykman_cli, transport, info, "12345679", "test-name-sym"
        )
        creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "test-name-sym" in creds

    def test_import_credential_symmetric_generate(
        self, ykman_cli, transport, info, management_key
    ):
        output = ykman_cli(
            "hsmauth",
            "credentials",
            "symmetric",
            "test-name-sym-gen",
            "-c",
            "12345679",
            "-g",
            "-m",
            management_key,
        ).output
        self.verify_credential_password(
            ykman_cli, transport, info, "12345679", "test-name-sym-gen"
        )
        assert "Generated ENC and MAC keys" in output

    def test_import_credential_symmetric_derived(
        self, ykman_cli, transport, info, management_key
    ):
        self.import_key_derived(
            ykman_cli,
            management_key,
            "test-name-sym-derived",
            "12345679",
        )
        self.verify_credential_password(
            ykman_cli, transport, info, "12345679", "test-name-sym-derived"
        )
        creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "test-name-sym-derived" in creds

    @condition.min_version(5, 6)
    def test_import_credential_asymmetric(self, ykman_cli, management_key):
        pair = generate_pem_eccp256_keypair()
        ykman_cli(
            "hsmauth",
            "credentials",
            "import",
            "test-name-asym",
            "-c",
            "12345679",
            "-m",
            management_key,
            "-",
            input=pair[0],
        )
        creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "test-name-asym" in creds

        public_key_exported = ykman_cli(
            "hsmauth", "credentials", "export", "test-name-asym", "-"
        ).stdout_bytes
        assert pair[1] == public_key_exported

    @condition.min_version(5, 6)
    def test_generate_credential_asymmetric(self, ykman_cli, management_key):
        ykman_cli(
            "hsmauth",
            "credentials",
            "generate",
            "test-name-asym-generated",
            "-c",
            "12345679",
            "-m",
            management_key,
        )

        creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "test-name-asym-generated" in creds

    def test_import_credential_touch_required(self, ykman_cli, management_key):
        self.import_key_derived(
            ykman_cli,
            management_key,
            "test-name-touch",
            touch_required=True,
        )

        creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "On" in creds
        assert "test-name-touch" in creds

    @condition.min_version(5, 6)
    def test_export_public_key_to_file(
        self, ykman_cli, management_key, eccp256_keypair, tmp_file
    ):
        private_key_file, public_key = eccp256_keypair
        ykman_cli(
            "hsmauth",
            "credentials",
            "import",
            "test-name-asym",
            "-c",
            "12345679",
            "-m",
            management_key,
            private_key_file,
        )

        ykman_cli(
            "hsmauth",
            "credentials",
            "export",
            "test-name-asym",
            tmp_file.name,
        )

        public_key_from_file = tmp_file.read()
        assert public_key_from_file == public_key

    @condition.min_version(5, 6)
    def test_export_public_key_symmetric_credential(self, ykman_cli, management_key):
        self.import_key_derived(
            ykman_cli,
            management_key,
            "test-name-sym",
        )

        with pytest.raises(SystemExit):
            ykman_cli("hsmauth", "credentials", "export", "test-name-sym")

    def test_delete_credential(self, ykman_cli, management_key):
        self.import_key_derived(ykman_cli, management_key, "delete-me")
        old_creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "delete-me" in old_creds
        ykman_cli(
            "hsmauth", "credentials", "delete", "delete-me", "-f", "-m", management_key
        )
        new_creds = ykman_cli("hsmauth", "credentials", "list").output
        assert "delete-me" not in new_creds

    @condition.min_version(5, 8)
    def test_change_credential_password(self, ykman_cli, management_key):
        credential_password = "123456789"
        new_credential_password = "987654321"
        label = "test-change-password"
        self.import_key_derived(
            ykman_cli, management_key, label, credential_password=credential_password
        )

        with pytest.raises(SystemExit):
            # Should fail - wrong current password
            ykman_cli(
                "hsmauth",
                "credentials",
                "change-password",
                label,
                "-c",
                "wrongvalue",
                "-n",
                new_credential_password,
            )

        # Should succeed
        ykman_cli(
            "hsmauth",
            "credentials",
            "change-password",
            label,
            "-c",
            credential_password,
            "-n",
            new_credential_password,
        )

    @condition.min_version(5, 8)
    def test_change_credential_password_admin(self, ykman_cli, management_key):
        credential_password = "123456789"
        new_credential_password = "987654321"
        label = "test-change-password"
        self.import_key_derived(
            ykman_cli,
            management_key,
            label,
            credential_password=credential_password,
        )

        with pytest.raises(SystemExit):
            # Should fail - wrong management key
            ykman_cli(
                "hsmauth",
                "credentials",
                "change-password",
                label,
                "-m",
                NON_DEFAULT_MANAGEMENT_KEY,
                "-n",
                new_credential_password,
            )

        # Should succeed
        ykman_cli(
            "hsmauth",
            "credentials",
            "change-password",
            label,
            "-m",
            management_key,
            "-n",
            new_credential_password,
        )


class TestManagementKey:
    def test_change_management_password(self, ykman_cli, management_key):
        ykman_cli(
            "hsmauth",
            "access",
            "change-management-password",
            "-m",
            management_key,
            "-n",
            NON_DEFAULT_MANAGEMENT_KEY,
        )

        with pytest.raises(SystemExit):
            # Should fail - wrong current key
            ykman_cli(
                "hsmauth",
                "access",
                "change-management-password",
                "-m",
                management_key,
                "-n",
                NON_DEFAULT_MANAGEMENT_KEY,
            )

        # Should succeed
        ykman_cli(
            "hsmauth",
            "access",
            "change-management-password",
            "-m",
            NON_DEFAULT_MANAGEMENT_KEY,
            "-n",
            NON_DEFAULT_MANAGEMENT_KEY,
        )

    @condition.check(lambda info: not info.pin_complexity, "PIN complexity")
    @condition.max_version(5, 7, 99)
    def test_change_management_key_generate(self, version, ykman_cli, management_key):
        if len(management_key) != 32:
            pytest.skip("string management key")

        output = ykman_cli(
            "hsmauth",
            "access",
            "change-management-key",
            "-m",
            management_key,
            "-g",
        ).output

        gen_key = re.search(r"[a-f0-9]{32}", output).group(0)

        ykman_cli(
            "hsmauth",
            "access",
            "change-management-password",
            "-m",
            gen_key,
            "-n",
            NON_DEFAULT_MANAGEMENT_KEY,
        )



#  vim: set fileencoding=utf-8 :

import re

import pytest
from ykman import __version__ as version
from ykman.otp import format_oath_code, generate_static_pw, time_challenge
from ykman.util import (
    _parse_pkcs12,
    is_pem,
    is_pkcs12,
    parse_certificates,
    parse_private_key,
)
from yubikit.core import InvalidPinError, Tlv, bytes2int
from yubikit.core.otp import modhex_decode, modhex_encode
from yubikit.management import FORM_FACTOR

from .util import open_file


def test_invalid_pin_exception_value_error():
    # Fail if InvalidPinError still inherits ValueError in ykman 6.0
    if int(version.split(".")[0]) != 5:
        assert not isinstance(InvalidPinError(3), ValueError)


def test_bytes2int():
    assert bytes2int(b"\x57") == 0x57
    assert bytes2int(b"\x12\x34") == 0x1234
    assert bytes2int(b"\xca\xfe\xd0\x0d") == 0xCAFED00D


@pytest.mark.parametrize(
    ("payload", "digits", "expected"),
    [
        (b"\0" * 20, None, "000000"),
        (b"\0" * 20, 8, "00000000"),
        (b"\x00\xbc\x61\x4e" + b"\0" * 16, None, "345678"),
        (b"\x49\x96\x02\xd2" + b"\0" * 16, 8, "34567890"),
    ],
)
def test_format_oath_code(payload, digits, expected):
    if digits is None:
        assert format_oath_code(payload) == expected
    else:
        assert format_oath_code(payload, digits) == expected


def test_generate_static_pw():
    template = r"^[cbdefghijklnrtuvCBDEFGHIJKLNRTUV]{%d}$"
    for length in range(0, 38):
        pattern = re.compile(template % length)
        assert pattern.fullmatch(generate_static_pw(length)), (
            f"Length {length} failed regex check"
        )


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("", b""),
        ("dteffuje", b"\x2d\x34\x4e\x83"),
        (
            "hknhfjbrjnlnldnhcujvddbikngjrtgh",
            b"\x69\xb6\x48\x1c\x8b\xab\xa2\xb6\x0e\x8f\x22\x17\x9b\x58\xcd\x56",
        ),
    ],
)
def test_modhex_decode(value, expected):
    assert modhex_decode(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (b"", ""),
        (b"\x2d\x34\x4e\x83", "dteffuje"),
        (
            b"\x69\xb6\x48\x1c\x8b\xab\xa2\xb6\x0e\x8f\x22\x17\x9b\x58\xcd\x56",
            "hknhfjbrjnlnldnhcujvddbikngjrtgh",
        ),
    ],
)
def test_modhex_encode(value, expected):
    assert modhex_encode(value) == expected


def test_parse_tlvs():
    tlvs = Tlv.parse_list(b"\x00\x02\xd0\x0d\xa1\x00\xfe\x04\xfe\xed\xfa\xce")
    assert len(tlvs) == 3

    assert tlvs[0].tag == 0
    assert tlvs[0].length == 2
    assert tlvs[0].value == b"\xd0\x0d"

    assert tlvs[1].tag == 0xA1
    assert tlvs[1].length == 0
    assert tlvs[1].value == b""

    assert tlvs[2].tag == 0xFE
    assert tlvs[2].length == 4
    assert tlvs[2].value == b"\xfe\xed\xfa\xce"


@pytest.mark.parametrize(
    ("timestamp", "expected"),
    [
        (0, b"\0" * 8),
        (12345678, b"\x00\x00\x00\x00\x00\x06G\x82"),
        (1484223461.2644958, b"\x00\x00\x00\x00\x02\xf2\xeaC"),
    ],
)
def test_time_challenge(timestamp, expected):
    assert time_challenge(timestamp) == expected


def test_tlv():
    assert Tlv(b"\xfe\6foobar") == Tlv(0xFE, b"foobar")

    tlv1 = Tlv(b"\0\5hello")
    tlv2 = Tlv(0xFE, b"")
    tlv3 = Tlv(0x12, b"hi" * 200)

    assert tlv1 == b"\0\5hello"
    assert tlv2 == b"\xfe\0"
    assert tlv3 == b"\x12\x82\x01\x90" + b"hi" * 200
    assert tlv1 + tlv2 + tlv3 == b"\0\5hello\xfe\0\x12\x82\x01\x90" + b"hi" * 200


def test_is_pkcs12_type_check():
    with pytest.raises(TypeError):
        is_pkcs12(None)


@pytest.mark.parametrize(
    "filename",
    [
        "rsa_2048_key.pem",
        "rsa_2048_key_encrypted.pem",
        "rsa_2048_cert.pem",
    ],
)
def test_is_pkcs12_rejects_pem_material(filename):
    with open_file(filename) as fh:
        assert not is_pkcs12(fh.read())


def test_is_pkcs12_accepts_pkcs12_blobs():
    with open_file("rsa_2048_key_cert.pfx") as rsa_2048_key_cert_pfx:
        data = rsa_2048_key_cert_pfx.read()

    assert is_pkcs12(data)
    parse_private_key(data, None)
    parse_certificates(data, None)

    with open_file("rsa_2048_key_cert_encrypted.pfx") as encrypted_pfx:
        assert is_pkcs12(encrypted_pfx.read())


def test_parse_pkcs12():
    with open_file("rsa_2048_key_cert.pfx") as rsa_2048_key_cert_pfx:
        data = rsa_2048_key_cert_pfx.read()

    key, certs = _parse_pkcs12(data, None)
    assert key is not None
    assert len(certs) == 1


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (b"just a byte string", False),
        (None, False),
        ("rsa_2048_key.pem", True),
        ("rsa_2048_key_encrypted.pem", True),
        ("rsa_2048_cert.pem", True),
        ("rsa_2048_key_cert.pfx", False),
        ("rsa_2048_cert_metadata.pem", True),
        ("rsa_2048_key_cert_encrypted.pfx", False),
    ],
)
def test_is_pem(value, expected):
    if isinstance(value, str):
        with open_file(value) as fh:
            contents = fh.read()
    else:
        contents = value
    assert bool(is_pem(contents)) is expected


@pytest.mark.parametrize(
    ("code", "expected"),
    [
        (0x00, FORM_FACTOR.UNKNOWN),
        (0x01, FORM_FACTOR.USB_A_KEYCHAIN),
        (0x02, FORM_FACTOR.USB_A_NANO),
        (0x03, FORM_FACTOR.USB_C_KEYCHAIN),
        (0x04, FORM_FACTOR.USB_C_NANO),
        (0x05, FORM_FACTOR.USB_C_LIGHTNING),
        (0x99, FORM_FACTOR.UNKNOWN),
    ],
)
def test_form_factor_from_code(code, expected):
    assert FORM_FACTOR.from_code(code) == expected


def test_form_factor_from_code_rejects_invalid_type():
    with pytest.raises(ValueError):
        FORM_FACTOR.from_code("im a string")



import os

import pytest
from cryptography.hazmat.primitives import serialization

from yubikit.core import Tlv
from yubikit.piv import OBJECT_ID, SLOT

from ....util import generate_self_signed_certificate


class TestReadWriteObject:
    def test_write_read_preserves_ansi_escapes(self, ykman_cli, keys):
        red = b"\x00\x1b[31m"
        blue = b"\x00\x1b[34m"
        reset = b"\x00\x1b[0m"
        data = (
            b"Hello, "
            + red
            + b"red"
            + reset
            + b" and "
            + blue
            + b"blue"
            + reset
            + b" world!"
        )
        ykman_cli(
            "piv",
            "objects",
            "import",
            "-m",
            keys.mgmt,
            "0x5f0001",
            "-",
            input=data,
        )
        output_data = ykman_cli(
            "piv", "objects", "export", "0x5f0001", "-"
        ).stdout_bytes
        assert data == output_data

    def test_read_write_read_is_noop(self, ykman_cli, keys):
        data = os.urandom(32)

        ykman_cli(
            "piv",
            "objects",
            "import",
            hex(OBJECT_ID.AUTHENTICATION),
            "-",
            "-m",
            keys.mgmt,
            input=data,
        )

        output1 = ykman_cli(
            "piv", "objects", "export", hex(OBJECT_ID.AUTHENTICATION), "-"
        ).stdout_bytes
        assert output1 == data

        ykman_cli(
            "piv",
            "objects",
            "import",
            hex(OBJECT_ID.AUTHENTICATION),
            "-",
            "-m",
            keys.mgmt,
            input=output1,
        )

        output2 = ykman_cli(
            "piv", "objects", "export", hex(OBJECT_ID.AUTHENTICATION), "-"
        ).stdout_bytes
        assert output2 == data

    def test_read_write_certificate_as_object(self, ykman_cli, keys):
        with pytest.raises(SystemExit):
            ykman_cli("piv", "objects", "export", hex(OBJECT_ID.AUTHENTICATION), "-")

        cert = generate_self_signed_certificate()
        cert_bytes_der = cert.public_bytes(encoding=serialization.Encoding.DER)

        input_tlv = Tlv(0x70, cert_bytes_der) + Tlv(0x71, b"\0") + Tlv(0xFE, b"")

        ykman_cli(
            "piv",
            "objects",
            "import",
            hex(OBJECT_ID.AUTHENTICATION),
            "-",
            "-m",
            keys.mgmt,
            input=input_tlv,
        )

        output1 = ykman_cli(
            "piv", "objects", "export", hex(OBJECT_ID.AUTHENTICATION), "-"
        ).stdout_bytes
        output_cert_bytes = Tlv.parse_dict(output1)[0x70]
        assert output_cert_bytes == cert_bytes_der

        output2 = ykman_cli(
            "piv",
            "certificates",
            "export",
            hex(SLOT.AUTHENTICATION),
            "-",
            "--format",
            "DER",
        ).stdout_bytes
        assert output2 == cert_bytes_der



from binascii import b2a_hex

import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa

from ... import condition
from .util import NON_DEFAULT_MANAGEMENT_KEY


def _verify_cert(cert, pubkey):
    cert_signature = cert.signature
    cert_bytes = cert.tbs_certificate_bytes

    if isinstance(pubkey, rsa.RSAPublicKey):
        pubkey.verify(
            cert_signature,
            cert_bytes,
            padding.PKCS1v15(),
            cert.signature_hash_algorithm,
        )
    elif isinstance(pubkey, ec.EllipticCurvePublicKey):
        pubkey.verify(
            cert_signature, cert_bytes, ec.ECDSA(cert.signature_hash_algorithm)
        )
    else:
        raise ValueError("Unsupported public key value")


def not_roca(version):
    return not ((4, 2, 0) <= version < (4, 3, 5))


class TestNonDefaultMgmKey:
    @pytest.fixture(autouse=True)
    def set_mgmt_key(self, ykman_cli, keys):
        ykman_cli(
            "piv",
            "access",
            "change-management-key",
            "-P",
            keys.pin,
            "-m",
            keys.mgmt,
            "-n",
            NON_DEFAULT_MANAGEMENT_KEY,
        )

    def _test_generate_self_signed(self, ykman_cli, keys, slot, algo):
        pubkey_output = ykman_cli(
            "piv",
            "keys",
            "generate",
            slot,
            "-a",
            algo,
            "-m",
            NON_DEFAULT_MANAGEMENT_KEY,
            "-",
        ).output
        ykman_cli(
            "piv",
            "certificates",
            "generate",
            slot,
            "-m",
            NON_DEFAULT_MANAGEMENT_KEY,
            "-s",
            "subject-" + algo,
            "-P",
            keys.pin,
            "-",
            input=pubkey_output,
        )
        output = ykman_cli("piv", "certificates", "export", slot, "-").output
        cert = x509.load_pem_x509_certificate(output.encode(), default_backend())
        _verify_cert(cert, cert.public_key())
        fingerprint = b2a_hex(cert.fingerprint(hashes.SHA256())).decode("ascii")

        output = ykman_cli("piv", "info").output
        assert fingerprint in output

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9a_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9a", "RSA2048")

    def test_generate_self_signed_slot_9a_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9a", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9c_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9c", "RSA2048")

    def test_generate_self_signed_slot_9c_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9c", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9d_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9d", "RSA2048")

    def test_generate_self_signed_slot_9d_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9d", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9e_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9e", "RSA2048")

    def test_generate_self_signed_slot_9e_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9e", "ECCP256")

    def _test_generate_csr(self, ykman_cli, keys, slot, algo):
        subject_input = "subject-" + algo
        pubkey_output = ykman_cli(
            "piv",
            "keys",
            "generate",
            slot,
            "-a",
            algo,
            "-m",
            NON_DEFAULT_MANAGEMENT_KEY,
            "-",
        ).output
        csr_output = ykman_cli(
            "piv",
            "certificates",
            "request",
            slot,
            "-P",
            keys.pin,
            "-",
            "-",
            "-s",
            subject_input,
            input=pubkey_output,
        ).output
        csr = x509.load_pem_x509_csr(csr_output.encode("utf-8"), default_backend())
        subject_output = csr.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[
            0
        ].value

        assert subject_input == subject_output

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9a_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9a", "RSA2048")

    def test_generate_csr_slot_9a_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9a", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9c_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9c", "RSA2048")

    def test_generate_csr_slot_9c_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9c", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9d_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9d", "RSA2048")

    def test_generate_csr_slot_9d_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9d", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9e_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9e", "RSA2048")

    def test_generate_csr_slot_9e_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9e", "ECCP256")


class TestProtectedMgmKey:
    @pytest.fixture(autouse=True)
    def protect_mgmt_key(self, ykman_cli, keys):
        ykman_cli(
            "piv",
            "access",
            "change-management-key",
            "-p",
            "-P",
            keys.pin,
            "-m",
            keys.mgmt,
        )

    def _test_generate_self_signed(self, ykman_cli, keys, slot, algo):
        pubkey_output = ykman_cli(
            "piv", "keys", "generate", slot, "-a", algo, "-P", keys.pin, "-"
        ).output
        ykman_cli(
            "piv",
            "certificates",
            "generate",
            slot,
            "-P",
            keys.pin,
            "-s",
            "subject-" + algo,
            "-",
            input=pubkey_output,
        )
        output = ykman_cli("piv", "certificates", "export", slot, "-").output
        cert = x509.load_pem_x509_certificate(output.encode(), default_backend())
        _verify_cert(cert, cert.public_key())
        fingerprint = b2a_hex(cert.fingerprint(hashes.SHA256())).decode("ascii")

        output = ykman_cli("piv", "info").output
        assert fingerprint in output

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9a_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9a", "RSA2048")

    def test_generate_self_signed_slot_9a_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9a", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9c_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9c", "RSA2048")

    def test_generate_self_signed_slot_9c_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9c", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9d_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9d", "RSA2048")

    def test_generate_self_signed_slot_9d_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9d", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_self_signed_slot_9e_rsa2048(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9e", "RSA2048")

    def test_generate_self_signed_slot_9e_eccp256(self, ykman_cli, keys):
        self._test_generate_self_signed(ykman_cli, keys, "9e", "ECCP256")

    def _test_generate_csr(self, ykman_cli, keys, slot, algo):
        subject_input = "subject-" + algo
        pubkey_output = ykman_cli(
            "piv", "keys", "generate", slot, "-a", algo, "-P", keys.pin, "-"
        ).output
        csr_output = ykman_cli(
            "piv",
            "certificates",
            "request",
            slot,
            "-P",
            keys.pin,
            "-",
            "-",
            "-s",
            subject_input,
            input=pubkey_output,
        ).output
        csr = x509.load_pem_x509_csr(csr_output.encode("utf-8"), default_backend())
        subject_output = csr.subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)[
            0
        ].value

        assert subject_input == subject_output

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9a_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9a", "RSA2048")

    def test_generate_csr_slot_9a_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9a", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9c_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9c", "RSA2048")

    def test_generate_csr_slot_9c_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9c", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9d_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9d", "RSA2048")

    def test_generate_csr_slot_9d_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9d", "ECCP256")

    @condition.yk4_fips(False)
    @condition.check(not_roca)
    def test_generate_csr_slot_9e_rsa2048(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9e", "RSA2048")

    def test_generate_csr_slot_9e_eccp256(self, ykman_cli, keys):
        self._test_generate_csr(ykman_cli, keys, "9e", "ECCP256")



from binascii import a2b_hex

import pytest

from ykman.hsmauth import generate_random_management_key
from yubikit.hsmauth import (
    CREDENTIAL_PASSWORD_LEN,
    MAX_LABEL_LEN,
    _parse_credential_password,
    _parse_label,
    _password_to_key,
)


class TestHsmAuthFunctions:
    def test_generate_random_management_key(self):
        output1 = generate_random_management_key()
        output2 = generate_random_management_key()

        assert isinstance(output1, bytes)
        assert isinstance(output2, bytes)
        assert 16 == len(output1) == len(output2)

    def test_parse_credential_password(self):
        parsed_credential_password = _parse_credential_password("123456")

        assert (
            b"123456\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            == parsed_credential_password
        )

    def test_parse_credential_password_wrong_length(self):
        with pytest.raises(ValueError):
            _parse_credential_password(b"1" * (CREDENTIAL_PASSWORD_LEN + 1))

    def test_parse_label(self):
        parsed_label = _parse_label("Default key")

        assert isinstance(parsed_label, bytes)

    def test_parse_label_wrong_length(self):
        with pytest.raises(ValueError):
            _parse_label("1" * (MAX_LABEL_LEN + 1))

        with pytest.raises(ValueError):
            _parse_label("")

    def test_password_to_key(self):
        assert (
            a2b_hex("090b47dbed595654901dee1cc655e420"),
            a2b_hex("592fd483f759e29909a04c4505d2ce0a"),
        ) == _password_to_key("password")

    def test__password_to_key_utf8(self):
        assert (
            a2b_hex("f320972c667ba5cd4d35119a6b0271a1"),
            a2b_hex("f10050ca688e5a6ce62b1ffb0f6f6869"),
        ) == _password_to_key("κόσμε")

    def test_password_to_key_bytes_fails(self):
        with pytest.raises(AttributeError):
            _password_to_key(b"password")

        with pytest.raises(AttributeError):
            _password_to_key(a2b_hex("cebae1bdb9cf83cebcceb5"))



# Copyright (c) 2021 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from time import sleep
from typing import Generator

from smartcard.Exceptions import CardConnectionException, NoCardException

from yubikit.core import TRANSPORT
from yubikit.core.fido import FidoConnection
from yubikit.core.otp import OtpConnection
from yubikit.core.smartcard import SmartCardConnection
from yubikit.management import DeviceInfo
from yubikit.support import get_name, read_info

from .base import YkmanDevice
from .device import list_all_devices, scan_devices
from .pcsc import list_devices as list_ccid

"""
Various helpers intended to simplify scripting.

Add an import to your script:

  from ykman import scripting as s

Example usage:

  yubikey = s.single()
  print("Here is a YubiKey:", yubikey)


  print("Insert multiple YubiKeys")
  for yubikey in s.multi():
      print("You inserted {yubikey}")
  print("You pressed Ctrl+C, end of script")

"""


class ScriptingDevice:
    """Scripting-friendly proxy for YkmanDevice.

    This wrapper adds some helpful utility methods useful for scripting.
    """

    def __init__(self, wrapped, info):
        self._wrapped = wrapped
        self._info = info
        self._name = get_name(info, self.pid.yubikey_type if self.pid else None)

    def __getattr__(self, attr):
        return getattr(self._wrapped, attr)

    def __str__(self):
        serial = self._info.serial
        return f"{self._name} ({serial})" if serial else self._name

    @property
    def info(self) -> DeviceInfo:
        return self._info

    @property
    def name(self) -> str:
        return self._name

    def reinsert(self, reinsert_cb=None, event=None):
        self._wrapped.reinsert(reinsert_cb=reinsert_cb, event=event)

    def otp(self) -> OtpConnection:
        """Establish a OTP connection."""
        return self.open_connection(OtpConnection)

    def smart_card(self) -> SmartCardConnection:
        """Establish a Smart Card connection."""
        return self.open_connection(SmartCardConnection)

    def fido(self) -> FidoConnection:
        """Establish a FIDO connection."""
        return self.open_connection(FidoConnection)


YkmanDevice.register(ScriptingDevice)


def single(*, prompt=True) -> ScriptingDevice:
    """Connect to a YubiKey.

    :param prompt: When set, you will be prompted to
        insert a YubiKey.
    """
    pids, state = scan_devices()
    n_devs = sum(pids.values())
    if prompt and n_devs == 0:
        print("Insert YubiKey...")
    while n_devs == 0:
        sleep(1.0)
        pids, new_state = scan_devices()
        n_devs = sum(pids.values())
    devs = list_all_devices()
    if len(devs) == 1:
        return ScriptingDevice(*devs[0])
    raise ValueError("Failed to get single YubiKey")


def multi(
    *, ignore_duplicates: bool = True, allow_initial: bool = False, prompt: bool = True
) -> Generator[ScriptingDevice, None, None]:
    """Connect to multiple YubiKeys.


    :param ignore_duplicates: When set, duplicates are ignored.
    :param allow_initial: When set, YubiKeys can be connected
        at the start of the function call.
    :param prompt: When set, you will be prompted to
        insert a YubiKey.
    """
    state = None
    handled_serials: set[int | None] = set()
    pids, _ = scan_devices()
    n_devs = sum(pids.values())
    if n_devs == 0:
        if prompt:
            print("Insert YubiKeys, one at a time...")
    elif not allow_initial:
        raise ValueError("YubiKeys must not be present initially.")

    while True:  # Run this until we stop the script with Ctrl+C
        pids, new_state = scan_devices()
        if new_state != state:
            state = new_state  # State has changed
            serials = set()
            if len(pids) == 0 and None in handled_serials:
                handled_serials.remove(None)  # Allow one key without serial at a time
            for device, info in list_all_devices():
                serials.add(info.serial)
                if info.serial not in handled_serials:
                    handled_serials.add(info.serial)
                    yield ScriptingDevice(device, info)
            if not ignore_duplicates:  # Reset handled serials to currently connected
                handled_serials = serials
        else:
            try:
                sleep(1.0)  # No change, sleep for 1 second.
            except KeyboardInterrupt:
                return  # Stop waiting


def _get_reader(reader) -> YkmanDevice:
    readers = [d for d in list_ccid(reader) if d.transport == TRANSPORT.NFC]
    if not readers:
        raise ValueError(f"No NFC reader found matching filter: '{reader}'")
    elif len(readers) > 1:
        names = [r.fingerprint for r in readers]
        raise ValueError(f"Multiple NFC readers matching filter: '{reader}' {names}")
    return readers[0]


def single_nfc(reader="", *, prompt=True) -> ScriptingDevice:
    """Connect to a YubiKey over NFC.

    :param reader: The name of the NFC reader.
    :param prompt: When set, you will prompted to place
        a YubiKey on NFC reader.
    """
    device = _get_reader(reader)
    while True:
        try:
            with device.open_connection(SmartCardConnection) as connection:
                info = read_info(connection)
            return ScriptingDevice(device, info)
        except NoCardException:
            if prompt:
                print("Place YubiKey on NFC reader...")
                prompt = False
            sleep(1.0)


def multi_nfc(
    reader="", *, ignore_duplicates=True, allow_initial=False, prompt=True
) -> Generator[ScriptingDevice, None, None]:
    """Connect to multiple YubiKeys over NFC.

    :param reader: The name of the NFC reader.
    :param ignore_duplicates: When set, duplicates are ignored.
    :param allow_initial: When set, YubiKeys can be connected
        at the start of the function call.
    :param prompt: When set, you will be prompted to place
        YubiKeys on the NFC reader.
    """
    device = _get_reader(reader)
    prompted = False

    try:
        with device.open_connection(SmartCardConnection) as connection:
            if not allow_initial:
                raise ValueError("YubiKey must not be present initially.")
    except NoCardException:
        if prompt:
            print("Place YubiKey on NFC reader...")
            prompted = True
        sleep(1.0)

    handled_serials: set[int | None] = set()
    current: int | None = -1
    while True:  # Run this until we stop the script with Ctrl+C
        try:
            with device.open_connection(SmartCardConnection) as connection:
                info = read_info(connection)
            if info.serial in handled_serials or current == info.serial:
                if prompt and not prompted:
                    print("Remove YubiKey from NFC reader.")
                    prompted = True
            else:
                current = info.serial
                if ignore_duplicates:
                    handled_serials.add(current)
                yield ScriptingDevice(device, info)
                prompted = False
        except NoCardException:
            if None in handled_serials:
                handled_serials.remove(None)  # Allow one key without serial at a time
            current = -1
            if prompt and not prompted:
                print("Place YubiKey on NFC reader...")
                prompted = True
        except CardConnectionException:
            pass
        try:
            sleep(1.0)  # No change, sleep for 1 second.
        except KeyboardInterrupt:
            return  # Stop waiting
